@@ -27,15 +27,17 @@ from markdown.preprocessors import Preprocessor
 import functools
 import urllib
 import re
-import codecs
 import os
 from . import util
 import textwrap
+import time
 
 MI = 1024 * 1024  # mebibyte (MiB)
 DEFAULT_URL_SIZE = MI * 32
 DEFAULT_URL_TIMEOUT = 10.0  # in seconds
 DEFAULT_URL_REQUEST_HEADERS = {}
+DEFAULT_MAX_RETRIES = 3
+DEFAULT_BACKOFF_FACTOR = 2
 
 
 class SnippetMissingError(Exception):
@@ -74,7 +76,9 @@ class SnippetPreprocessor(Preprocessor):
         '''
     )
 
-    RE_SNIPPET_FILE = re.compile(r'(?i)(.*?)(?:(:[0-9]*)?(:[0-9]*)?|(:[a-z][-_0-9a-z]*)?)$')
+    RE_SNIPPET_FILE = re.compile(
+        r'(?i)(.*?)(?:((?::-?[0-9]*){1,2}(?:(?:,(?=[-0-9:])-?[0-9]*)(?::-?[0-9]*)?)*)|(:[a-z][-_0-9a-z]*))?$'
+    )
 
     def __init__(self, config, md):
         """Initialize."""
@@ -92,6 +96,8 @@ class SnippetPreprocessor(Preprocessor):
         self.url_timeout = config['url_timeout']
         self.url_request_headers = config['url_request_headers']
         self.dedent_subsections = config['dedent_subsections']
+        self.max_retries = config['max_retries']
+        self.backoff_factor = config['backoff_factor']
         self.tab_length = md.tab_length
         super().__init__()
 
@@ -188,41 +194,59 @@ class SnippetPreprocessor(Preprocessor):
         The most recently used files are kept in a cache until the next reset.
         """
 
-        http_request = urllib.request.Request(url, headers=self.url_request_headers)
-        timeout = None if self.url_timeout == 0 else self.url_timeout
-        with urllib.request.urlopen(http_request, timeout=timeout) as response:
+        retries = self.max_retries
 
-            # Fail if status is not OK
-            status = response.status if util.PY39 else response.code
-            if status != 200:
-                raise SnippetMissingError(f"Cannot download snippet '{url}'")
+        while True:
+            try:
+                http_request = urllib.request.Request(url, headers=self.url_request_headers)
+                timeout = None if self.url_timeout == 0 else self.url_timeout
+                with urllib.request.urlopen(http_request, timeout=timeout) as response:
+                    # Fail if status is not OK
+                    status = response.status if util.PY39 else response.code
 
-            # We provide some basic protection against absurdly large files.
-            # 32MB is chosen as an arbitrary upper limit. This can be raised if desired.
-            content = None
-            if "content-length" not in response.headers:
-                # we have to read to know if we went over the max, but never more than `url_max_size`
-                # where `url_max_size` == 0 means unlimited
-                content = response.read(self.url_max_size) if self.url_max_size != 0 else response.read()
-                content_length = len(content)
-            else:
-                content_length = int(response.headers["content-length"])
+                    if status != 200:
+                        raise SnippetMissingError(f"Cannot download snippet '{url}' (HTTP Error {status})")
 
-            if self.url_max_size != 0 and content_length >= self.url_max_size:
-                raise ValueError(f"refusing to read payloads larger than or equal to {self.url_max_size}")
+                    # We provide some basic protection against absurdly large files.
+                    # 32MB is chosen as an arbitrary upper limit. This can be raised if desired.
+                    content = None
+                    if "content-length" not in response.headers:
+                        # we have to read to know if we went over the max, but never more than `url_max_size`
+                        # where `url_max_size` == 0 means unlimited
+                        content = response.read(self.url_max_size) if self.url_max_size != 0 else response.read()
+                        content_length = len(content)
+                    else:
+                        content_length = int(response.headers["content-length"])
 
-            # Nothing to return
-            if content_length == 0:
-                return ['']
+                    if self.url_max_size != 0 and content_length >= self.url_max_size:
+                        raise ValueError(f"refusing to read payloads larger than or equal to {self.url_max_size}")
 
-            if content is None:
-                # content-length was in the header, so we did not read yet
-                content = response.read()
+                    # Nothing to return
+                    if content_length == 0:
+                        return ['']
 
-            # Process lines
-            return [l.decode(self.encoding) for l in content.splitlines()]
+                    if content is None:
+                        # content-length was in the header, so we did not read yet
+                        content = response.read()
 
-    def parse_snippets(self, lines, file_name=None, is_url=False):
+                    # Process lines
+                    last = content.endswith((b'\r', b'\n'))
+                    s_lines = [l.decode(self.encoding) for l in content.splitlines()]
+                    if last:
+                        s_lines.append('')
+                    return s_lines
+
+            except urllib.error.HTTPError as e:  # noqa: PERF203
+                # Handle rate limited error codes
+                if e.code == 429 and retries:
+                    retries -= 1
+                    wait = self.backoff_factor * (self.max_retries - retries)
+                    time.sleep(wait)
+                    continue
+                raise SnippetMissingError(f"Cannot download snippet '{url}' (HTTP Error {e.code})") from e
+
+
+    def parse_snippets(self, lines, file_name=None, is_url=False, is_section=False):
         """Parse snippets snippet."""
 
         if file_name:
@@ -257,6 +281,20 @@ class SnippetPreprocessor(Preprocessor):
                     continue
 
             elif not block:
+                if not is_section:
+                    # Check for section line, if present remove, if escaped, reformat it
+                    m2 = self.RE_SNIPPET_SECTION.match(line)
+                    if m2 and m2.group('escape'):
+                        line = (
+                            m2.group('pre') + m2.group('escape').replace(';', '', 1) + m2.group('inline_marker') +
+                            m2.group('section') + m2.group('post')
+                        )
+                        m2 = None
+
+                    # Found a section that must be removed
+                    if m2 is not None:
+                        continue
+
                 # Not in snippet, and we didn't find an inline,
                 # so just a normal line
                 new_lines.append(line)
@@ -284,26 +322,26 @@ class SnippetPreprocessor(Preprocessor):
                     continue
 
                 # Get line numbers (if specified)
-                end = None
-                start = None
+                end = []
+                start = []
                 section = None
                 m = self.RE_SNIPPET_FILE.match(path)
-                path = m.group(1).strip()
+                path = '' if m is None else m.group(1).strip()
                 # Looks like we have an empty file and only lines specified
                 if not path:
                     if self.check_paths:
                         raise SnippetMissingError(f"Snippet at path '{path}' could not be found")
                     else:
                         continue
-                ending = m.group(3)
-                if ending and len(ending) > 1:
-                    end = int(ending[1:])
-                starting = m.group(2)
-                if starting and len(starting) > 1:
-                    start = max(0, int(starting[1:]) - 1)
-                section_name = m.group(4)
-                if section_name:
-                    section = section_name[1:]
+                if m.group(2):
+                    for nums in m.group(2)[1:].split(','):
+                        span = nums.split(':')
+                        st = int(span[0]) if span[0] else None
+                        start.append(st if st is None or st < 0 else max(0, st - 1))
+                        en = int(span[1]) if len(span) > 1 and span[1] else None
+                        end.append(en)
+                elif m.group(3):
+                    section = m.group(3)[1:]
 
                 # Ignore path links if we are in external, downloaded content
                 is_link = path.lower().startswith(('https://', 'http://'))
@@ -323,34 +361,43 @@ class SnippetPreprocessor(Preprocessor):
 
                     if not url:
                         # Read file content
-                        with codecs.open(snippet, 'r', encoding=self.encoding) as f:
-                            s_lines = [l.rstrip('\r\n') for l in f]
-                            if start is not None or end is not None:
-                                s = slice(start, end)
-                                s_lines = self.dedent(s_lines[s]) if self.dedent_subsections else s_lines[s]
-                            elif section:
-                                s_lines = self.extract_section(section, s_lines)
+                        with open(snippet, 'r', encoding=self.encoding) as f:
+                            last = False
+                            s_lines = []
+                            for l in f:
+                                last = l.endswith(('\r', '\n'))
+                                s_lines.append(l.strip('\r\n'))
+                            if last:
+                                s_lines.append('')
                     else:
                         # Read URL content
                         try:
                             s_lines = self.download(snippet)
-                            if start is not None or end is not None:
-                                s = slice(start, end)
-                                s_lines = self.dedent(s_lines[s]) if self.dedent_subsections else s_lines[s]
-                            elif section:
-                                s_lines = self.extract_section(section, s_lines)
                         except SnippetMissingError:
                             if self.check_paths:
                                 raise
                             s_lines = []
 
+                    if s_lines:
+                        total = len(s_lines)
+                        if start and end:
+                            final_lines = []
+                            for sel in zip(start, end):
+                                s_start = util.clamp(total + sel[0], 0, total) if sel[0] and sel[0] < 0 else sel[0]
+                                s_end = util.clamp(total + 1 + sel[1], 0, total) if sel[1] and sel[1] < 0 else sel[1]
+                                final_lines.extend(s_lines[slice(s_start, s_end, None)])
+                            s_lines = self.dedent(final_lines) if self.dedent_subsections else final_lines
+                        elif section:
+                            s_lines = self.extract_section(section, s_lines)
+
                     # Process lines looking for more snippets
                     new_lines.extend(
                         [
                             space + l2 for l2 in self.parse_snippets(
                                 s_lines,
                                 snippet,
-                                is_url=url
+                                is_url=url,
+                                is_section=section is not None
                             )
                         ]
                     )
@@ -396,7 +443,11 @@ class SnippetExtension(Extension):
             'url_max_size': [DEFAULT_URL_SIZE, "External URL max size (0 means no limit)- Default: 32 MiB"],
             'url_timeout': [DEFAULT_URL_TIMEOUT, 'Defualt URL timeout (0 means no timeout) - Default: 10 sec'],
             'url_request_headers': [DEFAULT_URL_REQUEST_HEADERS, "Extra request Headers - Default: {}"],
-            'dedent_subsections': [False, "Dedent subsection extractions e.g. 'sections' and/or 'lines'."]
+            'dedent_subsections': [False, "Dedent subsection extractions e.g. 'sections' and/or 'lines'."],
+            'max_retries': [
+                DEFAULT_MAX_RETRIES, "Maximum number of retry attempts for rate-limited requests - Default: 3"
+            ],
+            'backoff_factor': [DEFAULT_BACKOFF_FACTOR, "Backoff factor for retry attempts - Default: 2"]
         }
 
         super().__init__(*args, **kwargs)
